//#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//#+                                                                  
//#+    #     # #     # #     # ######   #####
//#+    ##   ## #     # ##   ## #     # #     #
//#+    # # # # #     # # # # # #     # #
//#+    #  #  # #     # #  #  # ######   #####
//#+    #     # #     # #     # #             #
//#+    #     # #     # #     # #       #     #
//#+    #     #  #####  #     # #        #####
//#+    
//#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//#+
//#+    Mumps Interpreter                            
//#+
//#+    Copyright (C) MM  Kevin C. O'Kane
//#+
//#+    Kevin C. O'Kane
//#+    Computer Science Department
//#+    University of Northern Iowa
//#+    Cedar Falls, IA 50614-0507
//#+    Tel 319 273 7322
//#+    okane@cs.uni.edu
//#+    http://www.cs.uni.edu/~okane
//#+
//#+    This program is free software; you can redistribute it and/or modify
//#+    it under the terms of the GNU General Public License as published by
//#+    the Free Software Foundation; either version 2 of the License, or
//#+    (at your option) any later version.
//#+
//#+    This program is distributed in the hope that it will be useful,
//#+    but WITHOUT ANY WARRANTY; without even the implied warranty of
//#+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//#+    GNU General Public License for more details.
//#+
//#+    You should have received a copy of the GNU General Public License
//#+    along with this program; if not, write to the Free Software
//#+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//#+
//#+                                                              
//#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


//#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//#+   
//#+               #####    #####  #####   ######  ######  
//#+               #    #     #    #    #  #       #  
//#+               #####      #    #    #  #####   #####  
//#+               #    #     #    #####   #       #  
//#+               #    #     #    #   #   #       #  
//#+               #####      #    #    #  ######  ######   
//#+   
//#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*

/*  #pragma inline */

#include <stdio.h>
#include "memsize.h"
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <ctype.h>

#if SYSTEM==DOS
#include <mem.h>
#include <stdlib.h>
#include <unistd.h>
#include <io.h>
#include <dos.h>
#endif

#include <sys/types.h>

#if SYSTEM==UNIX||SYSTEM==OS2
#include <sys/file.h>
#endif

#define GBLBUF 100


#define MARK 0xff

#ifndef LOCK_EX
#include <unistd.h>
#endif

#define TRXLIMIT 60
#define STORE 1
#define FOUND 1
#define RETRIEVE 0
#define NEXT 2
#define XNEXT 8
#define KILL 3
#define INIT 6
#define OPEN 7
#define EMPTY 5
#define CLOSE 4
#define NEXTMARK 1
#define PREVIOUS 9


        struct buffers {
                char *gblbf;
                long gblbfx;
                int gblmod;
                struct buffers *next;
                };

        static struct buffers bfr[GBLBUF];
        static struct buffers *gp0;
        static struct buffers *gp1;
        static struct buffers *gp3;
        static struct buffers *gp4;
        static struct buffers *gp5;
        static struct buffers *gproot;

        static int gblfd,gbldat;
        static gblmax,gblroot;
        static int trx;
        static int ReadOnly=1;

#if USYSTEM==LINUX
long tell(int);
#endif

void freadx(unsigned char **buf, long r, int flag);
void keyfix(unsigned char *);

int global(short g, unsigned char key[],unsigned char bd[]) {

        extern int errno;
		extern cgi_flag;
        long allocblk();
        char *calloc();
        static long trace[TRXLIMIT],ilong,zero=0,i,k,m,datptr;
        static long *root,*lastrec,*RDFlg;
        static short rtf,datlen,nxtln,tmp1,j,is,ks,ls,kk,ii;
        static short opnflg=0;
        static short tracep[TRXLIMIT];
        static unsigned char *block,*block1,*ptr1;
        static unsigned char bufblk[2048];
        static unsigned char prev[512];
        FILE *cfg;

#if SYSTEM!=UNIX
        static char cfgdata[]=UDAT;
        static char cfgkey[]=UKEY;
#endif

#if SYSTEM==UNIX
        static char cfgdata[]=UDAT;
        static char cfgkey[]=UKEY;
#endif

        rtf=1;  /* retry flag - prevents nxtln from being reset */
		prev[0]=0;

        keyfix(key);

        if (g==NEXT||g==XNEXT||g==PREVIOUS) { /* prepare NEXT cases */

                for (j=0; key[j]!=0; j++);

                if     (key[j-1]==1      /* $next -1 case */
                        && key[j-2]=='1'
                        && key[j-3]=='-'
                        && key[j-4]==1) {

                                if (g==PREVIOUS) key[j-3]='~';
								else key[j-3]=2;
                                key[j-2]=1;
                                key[j-1]=0;
                                j--;
                                }
                else

                if (g==NEXT) {
                        key[j-1]=2;
                        key[j]=1;
                        key[j+1]=0;
                        }

                if (rtf) {
                        for(nxtln=j-2; nxtln>0&&key[nxtln]!=NEXTMARK; nxtln--);
                        if (nxtln>0) nxtln++;
                        }
                rtf=0;
                goto srch;
                }

        if (g<4) goto srch; /* retrieve, store, kill */

        if (opnflg==0) { /* auto open */

                opnflg=1;

                for (i=0; i<GBLBUF; i++) { /* allocate buffers */
                        bfr[i].gblbf=(char *) malloc(1024);

                        if (bfr[i].gblbf==NULL)  {
                                printf("System memory error - buffers\n");
                                abort();
                                }

                        bfr[i].gblbfx= -1L;
                        bfr[i].gblmod=0;
                        }

                gblmax=i-1; 
                gblroot=gblmax;  /* buffer assignment */
                gp0=&bfr[0];

                for (i=0; i<=gblmax-2; i++) bfr[i].next=&bfr[i+1];

                bfr[gblmax-1].next=NULL;
                gproot=&bfr[gblroot];


#if UseConfigDB==1
/*-----------------------------------
	determine file system address
------------------------------------*/

		cfg=fopen("mps.cfg","r");

		if (cfg!=NULL) {

			while (fgets((char *)bufblk,80,cfg)!=NULL) {

				while (bufblk[strlen((char *)bufblk)-1]<' ') 
					bufblk[strlen((char *)bufblk)-1]=0;

				if (strncmp((char *)bufblk,"DATA",4)==0)
					strcpy((char *)cfgdata,(char *)&bufblk[5]);
				if (strncmp((char *)bufblk,"KEY",3)==0)
					strcpy((char *)cfgkey,(char *)&bufblk[4]);

				}
			fclose(cfg);
			}
#endif


reopen:
#if SYSTEM==UNIX
                gblfd=open(cfgkey,O_RDWR|O_EXCL);
                gbldat=open(cfgdata,O_RDWR|O_EXCL);
#else
                gblfd=open(cfgkey,O_RDWR|O_EXCL|O_BINARY);
                gbldat=open(cfgdata,O_RDWR|O_EXCL|O_BINARY);

		if (gblfd!=-1) while (lock(gblfd,0UL,1000UL)) sleep(1);
#endif

/*-----------------------------------------------
	unix file locking
-----------------------------------------------*/

#if SYSTEM!=DOS

#ifdef LOCK_EX
                flock(gblfd,LOCK_EX);  /* do not proceed until locked */
#endif
#ifndef LOCK_EX
                lockf(gblfd,F_LOCK,0L);
#endif
#endif

/*-----------------------------------------------
	end unix file locking
-----------------------------------------------*/

trunc:			if (gblfd==-1) { /* Does not exist - create and initialize */

#if SYSTEM==UNIX
                        gblfd=open(cfgkey,O_WRONLY|O_CREAT|O_TRUNC,
                                S_IRUSR|S_IWUSR);
                        gbldat=open(cfgdata,O_WRONLY|O_CREAT|O_TRUNC,
                                S_IRUSR|S_IWUSR);
		    if (gblfd==-1||gbldat==-1) {
				if (cgi_flag) printf("Content-type: text/html \n\n<html><body>");
				printf("Unable to open file system.\n\n");
				if (cgi_flag) printf("</body></html>\n\n");
				abort();
		    		}
#endif
#if SYSTEM==OS2||SYSTEM==DOS
                                
                        gblfd=open(cfgkey,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,
                                0777);
                        gbldat=open(cfgdata,O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,
                                0777);
		    	if (gblfd==-1||gbldat==-1) {
				if (cgi_flag) printf("Content-type: text/html \n\n<html><body>");
				printf("Unable to open file system.\n\n");
				if (cgi_flag) printf("</body></html>\n\n");
				abort();
		    		}
#endif


                        freadx(&block,0L,1); /* get a buffer address */

			      { int i; for (i=0; i<1024; i++) block[i]=0; }

                        for (is=0; is<=4; is++) { /* init to zeros */
                                lseek(gblfd,is*1024L,SEEK_SET);
                                write(gblfd,block,1024);
                                }

                        lseek(gbldat,0L,SEEK_SET); /* move to origin */
                        write(gbldat,block,2); /* init data file */

                        freadx((unsigned char **) &root,12L,1); 
                        *root= 777L; /* integrity flg */

                        freadx((unsigned char **) &root,0L,1); 
                        *root= 0L; /* root */

                        freadx((unsigned char **) &root,4L,1); 
                        *root= 0L; /* no meaning */

                        freadx((unsigned char **) &lastrec,8L,1); 
                        *lastrec= 1024L; /* extent */

                        lseek(gblfd,0L,SEEK_SET);
                        write(gblfd,bfr[gblroot].gblbf,1024); /* root block */

#if SYSTEM==OS2||SYSTEM==DOS
			unlock(gblfd,0UL,1000UL);
#endif

                        close(gblfd); 
			      close(gbldat);
			      chmod(cfgdata,0777);
			      chmod(cfgkey,0777);

                        goto reopen;
                        }

                freadx((unsigned char **) &root,12L,1); /* integrity flag */

                if (*root!=777L) {
				if (cgi_flag) printf("Content-type: text/html \n\n<html><body>");
                        printf("\nGlobal Array File Error - the database files are corrupt.\n");
				printf("You must Delete and then Restore the database files.\n");
				if (cgi_flag) printf("</body></html>\n\n");
                        close(gblfd); 
				close(gbldat);
				chmod(cfgdata,0777);
				chmod(cfgkey,0777);
                        abort();
                        }

                /*..........................................
                now done only when file system altered .....
                *root=0L; 
                lseek(gblfd,0L,SEEK_SET);
                write(gblfd,bfr[gblroot].gblbf,1024); 
                ............................................*/
				}

                if (g==INIT) {

#if SYSTEM==OS2||SYSTEM==DOS
					unlock(gblfd,0UL,1000UL);
#endif

#if SYSTEM!=DOS
#ifdef LOCK_EX
                	flock(gblfd,LOCK_UN);  /* unlock */
#endif
#endif
                	close(gblfd); 
                	close(gbldat);
			for (i=0; i<GBLBUF; i++) {
                 		bfr[i].gblbfx= -1L;
                 		bfr[i].gblmod=0;
                        }

			gblmax=i-1; 
			gblroot=gblmax;  /* buffer assignment */
			gp0=&bfr[0];
			for (i=0; i<=gblmax-2; i++) bfr[i].next=&bfr[i+1];
			bfr[gblmax-1].next=NULL;
			gproot=&bfr[gblroot];

			opnflg=1;
			gblfd=-1;
			g=OPEN;
			goto trunc;
                	}

        if (g==OPEN) return(1);

        if (g==CLOSE) {

                for (is=0; is<gblmax; is++) { /* dump global buffers */
                        if (bfr[is].gblmod>0) {
                                lseek(gblfd,bfr[is].gblbfx,SEEK_SET);
                                write(gblfd,bfr[is].gblbf,1024); }
                                }
                freadx((unsigned char **) &root,12L,1); 
                *root=777L; /* mark as proper close */

                lseek(gblfd,0L,SEEK_SET);
                write(gblfd,bfr[gblroot].gblbf,1024); /* dump super block */

#if SYSTEM==OS2||SYSTEM==DOS
		unlock(gblfd,0UL,1000UL);
#endif
#if SYSTEM!=DOS

#ifdef LOCK_EX
                flock(gblfd,LOCK_UN);  /* unlock */
#endif
#endif

                close(gblfd); 
                close(gbldat);

                return(1);
                }

        if (g==EMPTY) { /* count the empties */
                /* count empties */
                freadx((unsigned char **) &root,8L,0);
                lseek(gbldat,0L,SEEK_END);
                datptr=tell(gbldat);
                lseek(gblfd,0L,SEEK_END);
                k=tell(gblfd);
                i= (*root);
                is=0; for (j=0; j<GBLBUF; j++) if (bfr[j].gblbfx>0L) is++;
                sprintf(bd,
                " root=%ld data=%ld key=%ld buffers=%d Ver=29-Jan-2000",
                i,datptr,k,is);
                return(1); 
                }

srch:   freadx((unsigned char **) &root,0L,0);

        if (*root==0) { /*no tree*/

		if (g==NEXT||g==PREVIOUS) {strcpy(bd,"-1"); return(0);}

                if (g==RETRIEVE) {bd[0]=0; return(0); }

                if (g!=STORE) {bd[0]=0; return(0); }

                if (g==STORE) { /*insert case*/

                if (ReadOnly) {
                        freadx((unsigned char **) &RDFlg,12L,1); /* integrity flag */
                        *RDFlg=0L; /* mark as open - reset at close */
                        lseek(gblfd,0L,SEEK_SET);
                        write(gblfd,bfr[gblroot].gblbf,1024); /* dump super block */
                        ReadOnly=0;
                        }

                        k=allocblk();
                        freadx(&block,0L,1);
                        memcpy(block,&k,4);
                        freadx(&block,k,1); /* k block */
			for (j=0; j<1024; j++) block[j]=0;
                        j=4; strcpy(&block[j],key); /*key*/
                        j+=strlen(key)+1;
                        if (strlen(bd)==0) {
                                datptr=0L;
                                }
                        else {
                                datptr=lseek(gbldat,0L,SEEK_END);
                                datlen=strlen(bd)+1;
                                write(gbldat,&datlen,2); /* length code */
                                write(gbldat,bd,datlen); /* data */
                                }
                        memcpy(&block[j],&datptr,4); /* data address */
                        j+=4;
                        block[j++]=0;
                        memcpy(&block[j],&zero,4);
                        block[j+4]=MARK; block[j+5]=0;
                        return(1); }
                }

        /* search for key */

        i= *root; trx= -1;

nxtblk: trace[++trx]=i;
        freadx(&block,i,0);
        ptr1=block;

nxtkey: ptr1+=4;

#if SYSTEM==DOS
        while ( (is=_fstrcmp(key,ptr1) )>0) 
#else
        while ( (is=strcmp(key,ptr1) )>0) 
#endif
			{
				if (g==PREVIOUS) {
                    memcpy(&datptr,ptr1+strlen(ptr1)+1,4);
					if (datptr>=0L) strcpy(prev,ptr1); 
					}

                ptr1=ptr1+strlen(ptr1)+10; /*key & data*/
			}

        tracep[trx]=(short)(ptr1-block)-4;

        if (is<0) {
/*              ilong=*(long*)(ptr1-4);  */
                memcpy(&ilong,ptr1-4,4);  
                if (ilong<=0L) {
                        j=(short)(ptr1-block);
                        goto fail;
                        }
                i=ilong;
                goto nxtblk; 
                }

         /* found */

p000:			if (g==PREVIOUS) { 

						nxtln=strlen(key)-2;
						for(; key[nxtln]!=0&&key[nxtln]!=NEXTMARK; nxtln--);

                        if (strncmp(prev,key,nxtln)==0) {
							is=0; ls=nxtln+1;
							while(prev[ls]!=NEXTMARK) bd[is++]=prev[ls++];
							bd[is]=0;
							return(1);
                            }
                         else {
                                strcpy(bd,"-1");
                                return(0);
                                }
					} 

                if (g==RETRIEVE) { /*search operation*/

                        ptr1+=strlen(ptr1)+1;
/*                      datptr=*(long*)ptr1; */
                        memcpy(&datptr,ptr1,4);
                        if (datptr<0L)
                                {
                                bd[0]=0;
                                return(0); /* deleted */
                                }
                        if (datptr==0L) {
                                bd[0]=0;
                                return(1);
                                }
                        else {
                                lseek(gbldat,datptr,SEEK_SET);
                                read(gbldat,&datlen,2); /* length of data */
                                read(gbldat,bd,datlen); /* data */
                                }

                        return(1); }

                if (g==STORE) /* store */ {

                if (ReadOnly) {
                        freadx((unsigned char **) &RDFlg,12L,1); /* integrity flag */
                        *RDFlg=0L; /* mark as open - reset at close */
                        lseek(gblfd,0L,SEEK_SET);
                        write(gblfd,bfr[gblroot].gblbf,1024); /* dump super block */
                        ReadOnly=0;
                        }

                        freadx(&block,i,1);
                        ptr1+=strlen(ptr1)+1;

                        /*------------------------------*/
                        /* add new record - reuse old   */
                        /*------------------------------*/

                        memcpy(&datptr,ptr1,4);

			is=strlen(bd);

                        if (datptr<=0L) goto newrec;

                        lseek(gbldat,datptr,SEEK_SET);
                        read(gbldat,&datlen,2); /* length code */

                        if ( datlen > is ) {
                                lseek(gbldat,datptr,SEEK_SET);
                                goto save;
                                }

                        /*------------------------*/
                        /* delete old data record */
                        /* add new record at EOF  */
                        /*------------------------*/

newrec:                 if (is==0) {
                                datptr=0L;
                                goto save1;
                                }

                        datptr=lseek(gbldat,0L,SEEK_END);
                        datlen=is+1;

save:                   write(gbldat,&datlen,2); /* length code */
                        write(gbldat,bd,datlen); /* data */
save1:                  memcpy(ptr1,&datptr,4); /* data address */
                        return(1);
                        }

        j=(short)(ptr1-block);

                if (g==KILL) /* delete */ {

                if (ReadOnly) {
                        freadx((unsigned char **) &RDFlg,12L,1); /* integrity flag */
                        *RDFlg=0L; /* mark as open - reset at close */
                        lseek(gblfd,0L,SEEK_SET);
                        write(gblfd,bfr[gblroot].gblbf,1024); /* dump super block */
                        ReadOnly=0;
                        }

                        freadx(&block,i,1);
                        strcpy(key,ptr1); /* actual key deleted */
                        ptr1+=strlen(ptr1)+1;
                        datptr= -1L; /* mark as deleted - negative datptr */
                        memcpy(ptr1,&datptr,4);
                        return(1);
                        }

        if (g==NEXT||g==XNEXT) { /* found next */
                j+=strlen(&block[j])+10; /* next key */
                memcpy(&ilong,&block[j-4],4); /* look for lessthan childrn */
                if (ilong<=0L) {
                        goto fail;
                        }
                i=ilong;
                tracep[trx]=j-4; /* new parent */
                goto nxtblk;
                }

                /*end found*/

fail:		if (g==PREVIOUS) goto p000;

			if (g==NEXT||g==XNEXT) {

                if (block[j]!=MARK) {

                        if (g==XNEXT) {
                                /* modify key */
                                is=0; ls=j;
                                while (block[j]!=0) {
                                        key[is]=block[j];
                                        is++;
                                        j++;
                                        }
                                key[is]=0; j++;
                                memcpy(&datptr,&block[j],4); /* data ptr */
                                if (datptr<0L) {
                                        j=ls-4;
                                        ptr1=block+j;
                                        goto nxtkey; /* dltd key */
                                        }
                                is=1; ks=0;
                                bd[0]='^';
                                for (; block[ls]!=0; ls++) {
                                        if (block[ls]!=1) bd[is++]=block[ls];
                                        else
                                        if (ks) {
                                                bd[is++]='"';
                                                bd[is++]=',';
                                                bd[is++]='"';
                                                }
                                        else {
                                                ks=1; 
                                                bd[is++]='(';
                                                bd[is++]='"';
                                                }
                                        }
                                bd[is-1]=0;
                                bd[is-2]=')';
                                return(1);
                                }

                        /*-----------------
                        NEXT  case
                        ------------------*/

                        if (strncmp(&block[j],key,nxtln)==0)
                                {
                                /* modify key */
                                is=0; ls=j;
                                while (block[j]!=0) {
                                        key[is]=block[j];
                                        is++; j++;
                                        }
                                key[is]=0; j++;
                                memcpy(&datptr,&block[j],4);
                                if (datptr<0L) {
                                        j=ls-4;
                                        ptr1=block+j;
                                        goto nxtkey; /* dltd key */
                                        }
                                for(is=nxtln; key[is]!=NEXTMARK; is++);
                                key[is]=0;
                                strcpy(bd,&key[nxtln]);
				while (bd[0]==' ') strcpy(bd,&bd[1]);
                                return(1);
                                }
                            else
                                {
                                strcpy(bd,"-1");
                                return(0);
                                }
                        }

                /*----------------------
                end of block mark found
                ----------------------*/

upblock:        trx--;
                if (trx<0) {
                        strcpy(bd,"-1");
                        return(0);
                        }
                i=trace[trx];
                j=tracep[trx]+4;  /* key */
                freadx(&block,i,0);
                if (block[j]==MARK) goto upblock;
                goto fail; /* must be parent */
                }

        if (g==STORE) {

        /* store new key */

        k=0L;
        if (!bd[0]) datptr=0L; /* zero length data field */
        else {
                datptr=lseek(gbldat,0L,SEEK_END);
                datlen=strlen(bd)+1;
                write(gbldat,&datlen,2); /* length code */
                write(gbldat,bd,datlen); /* data */
                }

f1:     freadx(&block,i,1); /* marked for write */

        ls=j;

        while(block[ls]!=MARK) ls+=strlen(&block[ls])+10;

        ks=strlen(key);
        if (ks+ls>1000) goto split; /*no room*/

        /* simple insert */

        ii=(j-4)+ks+10; 
	  ls++;
/*      for (kk=ls+ks+10; kk>=ii; kk--) block[kk]=block[ls--]; */
        memmove(&block[(j-4)+ks+10],&block[j-4],ls-(j-4));  
        memcpy(&block[j-4],&k,4);          
        strcpy(&block[j],key);           
        j+=ks+1;
        memcpy(&block[j],&datptr,4);
        block[j+4]=0;

        return(1);

        /* split insert */
        /* insert prior to j-4 */

split:  ii=j-4;

        for(kk=0; kk<ii; kk++) bufblk[kk]=block[kk]; 

        memcpy(&bufblk[j-4],&k,4);          /* child */
        strcpy(&bufblk[j],key);             /* key */
        j+=ks+1;
        memcpy(&bufblk[j],&datptr,4);       /* data address */
        bufblk[j+4]=0;
        for(kk=j+5; ii<1024; kk++,ii++) bufblk[kk]=block[ii];
        for(; kk<2048; kk++) bufblk[kk]=0;

        /* make j forward new block and other things*/

        j=0;                                /* find break point */
        while (j<500) {
                j+=4;
                j+=strlen(&bufblk[j])+6;
                }

        j+=4; tmp1=j;
        strcpy(key,&bufblk[j]);             /* extract key to be sent up */
        j+=strlen(&bufblk[j])+1;            /* advance to data pointer */
        memcpy(&datptr,&bufblk[j],4);       /* extract data to be sent up */
        j+=5;                               /* advance past data pointer */

        freadx(&block,i,1);                 /* make memory resident */
        memcpy(block,&bufblk[j],1024);      /* right block */

        for (kk=tmp1; kk<2048; kk++)
                bufblk[kk]=0;               /* zap old right block
                                               but hold child ptr */

        bufblk[tmp1]=MARK;
        bufblk[tmp1+1]=0;                   /* new end of block */

        k=allocblk();                       /* allocate new left block */
        freadx(&block1,k,1);                /* make memory resident */
        memcpy(block1,bufblk,1024);         /* left part of block */

        /* fix parent */

        trx--; if (trx<0) goto rootfix;
        i=trace[trx]; j=tracep[trx]+4;
        goto f1;

rootfix:
        m=allocblk();
        freadx(&block,0L,1);
        memcpy(block,&m,4);
        freadx(&block,m,1);
        for(kk=0; kk<1024; kk++) block[kk]=0;
        memcpy(block,&k,4);
        strcpy(&block[4],key);
        j=5+strlen(key);
        memcpy(&block[j],&datptr,4);
        block[j+4]=0;
        j+=5;
        memcpy(&block[j],&trace[0],4);
        block[j+4]=MARK; block[j+5]=0;
        return(1);

        }

        bd[0]=0; return(0);  /* default failure */

        }

long allocblk()  {

long *lastrec;
long rec;
unsigned char *block;

freadx((unsigned char **) &lastrec,8L,1);
rec= *lastrec;
*lastrec+=1024L;
freadx(&block,(*lastrec),1);
return(rec);
}

void freadx(unsigned char **buf, long r, int flag)  {

if (r<1024L) { /* root block */

        if (gproot->gblbfx<0L) { /* load root */
                lseek(gblfd,0L,SEEK_SET);
                read(gblfd,gproot->gblbf,1024);
                gproot->gblbfx=0L;
                *buf=gproot->gblbf+r;
                return; 
                }

                *buf=gproot->gblbf+r;
                return; 
        }

gp1=gp0; 
gp3=NULL;
while (gp1!=NULL) {
        if (gp1->gblbfx==r) {
                if (flag) gp1->gblmod=1;
                *buf=gp1->gblbf;
                if (gp1==gp0) return;
                gp4=gp1->next;
                gp1->next=gp0;
                gp0=gp1;
                gp3->next=gp4;
                return;
                }
        gp5=gp3;
        gp3=gp1;
        gp1=gp1->next;
        }

if (gp3->gblmod) { /* write modified block out */
        lseek(gblfd,(long) gp3->gblbfx,SEEK_SET);
        write(gblfd,gp3->gblbf,1024);
        } 

lseek(gblfd,r,SEEK_SET);
read(gblfd,gp3->gblbf,1024);
*buf=gp3->gblbf;
gp3->gblbfx=r;
gp3->gblmod=flag;
gp3->next=gp0;
gp0=gp3;
gp5->next=NULL;
return; 
}

void keyfix(unsigned char * vxd) {

#include "externs.h"

int ls,is,ks,js,ms,ns;
char * p1;

p1=vxd;

/************************************************************************
*                                                                       *
*       remove leading ^.  insert current default directory.            *
*                                                                       *
************************************************************************/

        if (*vxd=='^') strcpy(vxd,vxd+1);

/************************************************************************
*                                                                       *
*       convert markers to 1's                                          *
*                                                                       *
************************************************************************/

	is=0;
        while (*vxd!=0) { 
		if (*vxd>127) *vxd=1; 
		vxd++;
		is++;
		}

/************************************************************************
*                                                                   
*       pad numeric subscripts for collating sequence compatibility  
*									
*		can be disabled with gpad setting				
*   
************************************************************************/

if (gpad) {
	ls=is;
        for (is=0; p1[is]!=0; is++) {
next_arg:       if (p1[is]!=1) continue; /* find next index */
                is++; 
				if (p1[is]==0) break; 
				ks=0;
                for (js=is; p1[js]!=0; js++) {
                        if (p1[js]==1) break; 
						ks++;
                        if (ks>8) goto next_arg;
                        if (isdigit(p1[js])) continue;
                                else goto next_arg; 
						}
                for (ms=ls+(8-ks); ms>is; ms--) p1[ms]=p1[ms-(8-ks)];
                ls=ls+(8-ks);
                for (ns=0; ns<=(7-ks); ns++) p1[is+ns]=' ';
                is=is+(8-ks); 
		} 
	}
return;
}

#if USYSTEM==LINUX
long tell(int i) { 
	FILE *p; 
	long j; 
	p=fdopen(i,"r"); 
	j=ftell(p); 
	fclose(p);
	return j; 
}
#endif
